from distutils.core import setup

from _setup_common import name, version, author, maintainer, maintainer_email, description, long_description, packages, data_files, license, classifiers

setup(
    name=name,
    version=version,
    author=author,
    maintainer=maintainer,
    maintainer_email=maintainer_email,
    description=description,
    long_description=long_description,
    packages=packages,
    data_files=data_files,
    license=license,
    classifiers=classifiers,
)